package de.ozgcloud.operator;

import java.util.Base64;

import org.apache.commons.collections.MapUtils;
import org.springframework.stereotype.Component;

import io.fabric8.kubernetes.api.model.Secret;
import io.javaoperatorsdk.operator.api.reconciler.Cleaner;
import io.javaoperatorsdk.operator.api.reconciler.Context;
import io.javaoperatorsdk.operator.api.reconciler.ControllerConfiguration;
import io.javaoperatorsdk.operator.api.reconciler.DeleteControl;
import io.javaoperatorsdk.operator.api.reconciler.Reconciler;
import io.javaoperatorsdk.operator.api.reconciler.UpdateControl;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@RequiredArgsConstructor
@ControllerConfiguration
@Component
public class ElasticsearchReconciler implements Reconciler<OzgCloudElasticsearchCustomResource>, Cleaner<OzgCloudElasticsearchCustomResource> {

	private final OzgCloudElasticsearchService service;

	@Override
	public UpdateControl<OzgCloudElasticsearchCustomResource> reconcile(OzgCloudElasticsearchCustomResource resource,
			Context<OzgCloudElasticsearchCustomResource> context) {
		try {
			var namespace = resource.getMetadata().getNamespace();
			LOG.info("{}: Reconcile user", namespace);
			var secret = service.getOrCreateCredentialSecret(resource, context);
			service.createIndexIfMissing(namespace);
			service.createSecurityRoleIfMissing(namespace);
			service.createSecurityUserIfMissing(namespace, getPassword(secret));
			service.createCertificateIfMissing(namespace);
			LOG.info("{}: Reconcile user successful.", namespace);
			return OzgCloudElasticsearchUpdateControlBuilder.fromResource(resource).withStatus(CustomResourceStatus.OK).build();
		} catch (Exception exception) {
			LOG.warn(resource.getMetadata().getNamespace() + ": Reconcile user failed.", exception);
			return buildExceptionUpdateControl(resource, exception);
		}
	}

	String getPassword(Secret secret) {
		return decode(MapUtils.getString(secret.getData(), OzgCloudElasticsearchSecretHelper.CREDENTIAL_SECRET_PASSWORD_FIELD));
	}

	private String decode(String encodedPassword) {
		try {
			return new String(Base64.getDecoder().decode(encodedPassword));
		} catch (Exception e) {
			throw new RuntimeException("Could not decode password from secret.");
		}
	}

	UpdateControl<OzgCloudElasticsearchCustomResource> buildExceptionUpdateControl(OzgCloudElasticsearchCustomResource resource,
			Exception exception) {
		return OzgCloudElasticsearchUpdateControlBuilder
				.fromResource(resource)
				.withStatus(CustomResourceStatus.ERROR)
				.withReschedule(OperatorConfig.RECONCILER_RETRY_SECONDS_ON_ERROR)
				.withMessage(exception.getMessage())
				.build();
	}

	@Override
	public DeleteControl cleanup(OzgCloudElasticsearchCustomResource resource, Context<OzgCloudElasticsearchCustomResource> context) {
		var namespace = resource.getMetadata().getNamespace();
		try {
			service.deleteSecurityUserIfExists(namespace);
			service.deleteSecurityRoleIfExists(namespace);
			service.deleteIndexIfExists(namespace);
			return DeleteControl.defaultDelete();
		} catch (Exception e) {
			LOG.warn(resource.getMetadata().getNamespace() + ": Could not cleanup elasticsearch resource.", e);
			return DeleteControl.defaultDelete();
		}
	}
}